from edc_lab import LabProfile
from edc_lab_panel.panels import (
    blood_glucose_panel,
    fbc_panel,
    hba1c_panel,
    hba1c_poc_panel,
    lft_panel,
    lipids_panel,
    rft_panel,
)

subject_lab_profile = LabProfile(
    name="subject_lab_profile",
    requisition_model="edc_metadata.subjectrequisition",
    reference_range_collection_name="my_reportables",
)

subject_lab_profile.add_panel(fbc_panel)
subject_lab_profile.add_panel(blood_glucose_panel)
subject_lab_profile.add_panel(hba1c_panel)
subject_lab_profile.add_panel(hba1c_poc_panel)
subject_lab_profile.add_panel(lipids_panel)
subject_lab_profile.add_panel(lft_panel)
subject_lab_profile.add_panel(rft_panel)



from django.db import models
from edc_constants.choices import YES_NO
from edc_lab.choices import RESULT_QUANTIFIER
from edc_lab.constants import EQ
from edc_reportable import PERCENT
from edc_reportable.choices import REPORTABLE


class Hba1cModelMixin(models.Model):
    is_poc = models.CharField(
        verbose_name="Was a point-of-care test used?",
        max_length=15,
        choices=YES_NO,
        null=True,
    )

    hba1c_value = models.DecimalField(
        verbose_name="HbA1c value",
        max_digits=8,
        decimal_places=4,
        null=True,
        blank=True,
    )

    hba1c_quantifier = models.CharField(
        max_length=10,
        choices=RESULT_QUANTIFIER,
        default=EQ,
        null=True,
        blank=True,
    )

    hba1c_units = models.CharField(
        verbose_name="units",
        max_length=15,
        default=PERCENT,
        null=True,
        blank=True,
    )

    hba1c_abnormal = models.CharField(
        verbose_name="abnormal", choices=YES_NO, max_length=25, null=True, blank=True
    )

    hba1c_reportable = models.CharField(
        verbose_name="reportable",
        choices=REPORTABLE,
        max_length=25,
        null=True,
        blank=True,
    )

    class Meta:
        abstract = True



from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from edc_lab_panel.model_mixin_factory import reportable_result_model_mixin_factory
from edc_reportable import GRAMS_PER_DECILITER, PERCENT
from edc_reportable.units import (
    CELLS_PER_MILLIMETER_CUBED,
    CELLS_PER_MILLIMETER_CUBED_DISPLAY,
    FEMTOLITERS_PER_CELL,
    PICOGRAMS_PER_CELL,
    TEN_X_9_PER_LITER,
)


class HaemoglobinModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="haemoglobin",
        verbose_name="Haemoglobin",
        units_choices=((GRAMS_PER_DECILITER, GRAMS_PER_DECILITER),),
        decimal_places=1,
    ),
    models.Model,
):
    class Meta:
        abstract = True


class HctModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="hct",
        verbose_name="Hematocrit",
        units_choices=((PERCENT, PERCENT),),
        validators=[MinValueValidator(1.0), MaxValueValidator(999.0)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class MchModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="mch",
        units_choices=((PICOGRAMS_PER_CELL, PICOGRAMS_PER_CELL),),
    ),
    models.Model,
):
    class Meta:
        abstract = True


class MchcModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="mchc",
        units_choices=((GRAMS_PER_DECILITER, GRAMS_PER_DECILITER),),
    ),
    models.Model,
):
    class Meta:
        abstract = True


class McvModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="mcv",
        units_choices=((FEMTOLITERS_PER_CELL, FEMTOLITERS_PER_CELL),),
    ),
    models.Model,
):
    class Meta:
        abstract = True


class NeutrophilModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="neutrophil",
        verbose_name="Neutrophil (abs)",
        units_choices=((TEN_X_9_PER_LITER, TEN_X_9_PER_LITER),),
        decimal_places=2,
        validators=[MinValueValidator(0.00), MaxValueValidator(9999)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class NeutrophilDiffModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="neutrophil_diff",
        verbose_name="Neutrophil (diff)",
        units_choices=((PERCENT, PERCENT),),
        decimal_places=2,
        validators=[MinValueValidator(0.00), MaxValueValidator(9999)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class LymphocyteModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="lymphocyte",
        verbose_name="Lymphocyte (abs)",
        units_choices=((TEN_X_9_PER_LITER, TEN_X_9_PER_LITER),),
        decimal_places=2,
        validators=[MinValueValidator(0.00), MaxValueValidator(9999)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class LymphocyteDiffModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="lymphocyte_diff",
        verbose_name="Lymphocyte (diff)",
        units_choices=((PERCENT, PERCENT),),
        decimal_places=2,
        validators=[MinValueValidator(0.00), MaxValueValidator(9999)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class PlateletsModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="platelets",
        verbose_name="Platelets",
        units_choices=(
            (TEN_X_9_PER_LITER, TEN_X_9_PER_LITER),
            (CELLS_PER_MILLIMETER_CUBED, CELLS_PER_MILLIMETER_CUBED_DISPLAY),
        ),
        decimal_places=0,
        validators=[MinValueValidator(1), MaxValueValidator(9999)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class RbcModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="rbc",
        verbose_name="Red blood cell count",
        units_choices=(
            (TEN_X_9_PER_LITER, TEN_X_9_PER_LITER),
            (CELLS_PER_MILLIMETER_CUBED, CELLS_PER_MILLIMETER_CUBED),
        ),
        validators=[MinValueValidator(1.0), MaxValueValidator(999999.0)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class WbcModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="wbc",
        units_choices=(
            (TEN_X_9_PER_LITER, TEN_X_9_PER_LITER),
            (CELLS_PER_MILLIMETER_CUBED, CELLS_PER_MILLIMETER_CUBED_DISPLAY),
        ),
    ),
    models.Model,
):
    class Meta:
        abstract = True



from copy import deepcopy

from django.apps import apps as django_apps
from django.test import TestCase
from edc_action_item import site_action_items
from edc_appointment.models import Appointment
from edc_constants.constants import NO, NOT_APPLICABLE, YES
from edc_lab.models import Panel
from edc_reportable import GRADE3, GRAMS_PER_DECILITER, PERCENT
from edc_utils import get_utcnow
from edc_visit_tracking.constants import SCHEDULED

from edc_lab_results.action_items import register_actions

from ..forms import BloodResultsFbcForm, BloodResultsHba1cForm
from ..test_case_mixin import TestCaseMixin


class TestBloodResultForm(TestCaseMixin, TestCase):
    def setUp(self):
        site_action_items.registry = {}
        register_actions()
        self.subject_identifier = self.enroll()
        appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code="1000",
        )
        self.subject_visit = django_apps.get_model(
            "edc_visit_tracking.subjectvisit"
        ).objects.create(
            report_datetime=get_utcnow(),
            appointment=appointment,
            reason=SCHEDULED,
        )
        fbc_panel = Panel.objects.get(name="fbc")
        requisition = django_apps.get_model("edc_metadata.subjectrequisition").objects.create(
            subject_visit=self.subject_visit,
            panel=fbc_panel,
            requisition_datetime=self.subject_visit.report_datetime,
        )
        self.data = dict(
            report_datetime=self.subject_visit.report_datetime,
            subject_visit=self.subject_visit,
            assay_datetime=self.subject_visit.report_datetime,
            requisition=requisition,
            action_identifier="-",
            results_reportable=NOT_APPLICABLE,
            results_abnormal=NO,
        )

    def test_fbc_ok(self):
        data = deepcopy(self.data)
        form = BloodResultsFbcForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)

    def test_missing_units(self):
        data = deepcopy(self.data)
        data.update(haemoglobin_value=10)
        form = BloodResultsFbcForm(data=data)
        form.is_valid()
        self.assertIn("haemoglobin_units", form._errors)

    def test_haemoglobin_abnormal_required(self):
        data = deepcopy(self.data)
        data.update(haemoglobin_value=10, haemoglobin_units=GRAMS_PER_DECILITER)
        form = BloodResultsFbcForm(data=data)
        form.is_valid()
        self.assertIn("haemoglobin_abnormal", form._errors)

    def test_haemoglobin_reportable_required(self):
        data = deepcopy(self.data)
        data.update(
            haemoglobin_value=10,
            haemoglobin_units=GRAMS_PER_DECILITER,
            haemoglobin_abnormal=NO,
        )
        form = BloodResultsFbcForm(data=data)
        form.is_valid()
        self.assertIn("haemoglobin_reportable", form._errors)

    def test_haemoglobin_normal(self):
        data = deepcopy(self.data)
        data.update(
            haemoglobin_value=14,
            haemoglobin_units=GRAMS_PER_DECILITER,
            haemoglobin_abnormal=NO,
            haemoglobin_reportable=NOT_APPLICABLE,
        )
        form = BloodResultsFbcForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)

    def test_haemoglobin_high(self):
        data = deepcopy(self.data)
        data.update(
            haemoglobin_value=18,
            haemoglobin_units=GRAMS_PER_DECILITER,
            haemoglobin_abnormal=YES,
            haemoglobin_reportable=NO,
            results_abnormal=YES,
            results_reportable=NO,
        )
        form = BloodResultsFbcForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)

    def test_haemoglobin_g4(self):
        data = deepcopy(self.data)
        data.update(
            haemoglobin_value=7.5,
            haemoglobin_units=GRAMS_PER_DECILITER,
            haemoglobin_abnormal=YES,
            haemoglobin_reportable=GRADE3,
            results_abnormal=YES,
            results_reportable=YES,
        )
        form = BloodResultsFbcForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)


class TestBloodResultFormForPoc(TestCaseMixin, TestCase):
    def setUp(self):
        super().setUp()
        site_action_items.registry = {}
        register_actions()
        self.subject_identifier = self.enroll()
        appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code="1000",
        )
        self.subject_visit = django_apps.get_model(
            "edc_visit_tracking.subjectvisit"
        ).objects.create(
            report_datetime=get_utcnow(),
            appointment=appointment,
            reason=SCHEDULED,
        )
        self.data = dict(
            report_datetime=self.subject_visit.report_datetime,
            subject_visit=self.subject_visit,
            assay_datetime=self.subject_visit.report_datetime,
            results_reportable=NOT_APPLICABLE,
            results_abnormal=NO,
        )

    def test_is_poc_does_not_require_requisition(self):
        data = deepcopy(self.data)

        data.update(is_poc=YES)
        form = BloodResultsHba1cForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)

        data.update(
            hba1c_value=5.0,
            hba1c_units=PERCENT,
            hba1c_abnormal=NO,
            hba1c_reportable=NOT_APPLICABLE,
        )
        form = BloodResultsHba1cForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)

        data.update(hba1c_value=4.3)
        form = BloodResultsHba1cForm(data=data)
        form.is_valid()
        self.assertIn("HBA1C is abnormal", str(form._errors.get("hba1c_value")))

        hba1c_panel = Panel.objects.get(name="hba1c")
        requisition = django_apps.get_model("edc_metadata.subjectrequisition").objects.create(
            subject_visit=self.subject_visit,
            panel=hba1c_panel,
            requisition_datetime=self.subject_visit.report_datetime,
        )

        data.update(requisition=requisition, hba1c_value=5.0)
        form = BloodResultsHba1cForm(data=data)
        form.is_valid()
        self.assertIn("This field is not required", str(form._errors.get("requisition")))

    def test_not_poc_requires_requisition(self):
        data = deepcopy(self.data)

        data.update(is_poc=NO)
        form = BloodResultsHba1cForm(data=data)
        form.is_valid()
        self.assertIn("This field is required", str(form._errors.get("requisition")))

        hba1c_panel = Panel.objects.get(name="hba1c")
        requisition = django_apps.get_model("edc_metadata.subjectrequisition").objects.create(
            subject_visit=self.subject_visit,
            panel=hba1c_panel,
            requisition_datetime=self.subject_visit.report_datetime,
        )

        data.update(requisition=requisition)
        form = BloodResultsHba1cForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)

        data.update(
            hba1c_value=5.0,
            hba1c_units=PERCENT,
            hba1c_abnormal=NO,
            hba1c_reportable=NOT_APPLICABLE,
        )
        form = BloodResultsHba1cForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)



from typing import Any

from django.apps import apps as django_apps
from django.db import models
from edc_constants.choices import YES_NO, YES_NO_NA
from edc_reportable.site_reportables import site_reportables
from edc_reportable.utils import get_reference_range_collection_name

from ..calculate_missing import calculate_missing


class BloodResultsFieldsModelMixin(models.Model):
    results_abnormal = models.CharField(
        verbose_name="Are any of the above results abnormal?",
        choices=YES_NO,
        max_length=25,
        help_text=(
            "Abnormal results present at baseline or continuing from baseline not included."
        ),
    )

    results_reportable = models.CharField(
        verbose_name="If any results are abnormal, are results within grade 3 or above?",
        max_length=25,
        choices=YES_NO_NA,
        help_text=(
            "If YES, this value will open Adverse Event Form. Grade 3 and 4 results "
            "present at baseline or continuing from baseline not included"
        ),
    )

    summary = models.TextField(null=True, blank=True)

    missing_count = models.IntegerField(
        default=0,
        editable=False,
        help_text="A count of fields left blank",
    )

    missing = models.TextField(
        null=True,
        editable=False,
        help_text="calculated string of field names that have been left blank",
    )

    class Meta:
        abstract = True


class BloodResultsMethodsModelMixin(models.Model):

    """Requires additional attrs `subject_visit` and `requisition`"""

    def save(self, *args, **kwargs):
        self.summary = "\n".join(self.get_summary())
        self.missing_count, self.missing = calculate_missing(self, self.lab_panel)
        super().save(*args, **kwargs)

    def get_summary_options(self: Any) -> dict:
        model_cls = django_apps.get_model("edc_registration.registeredsubject")
        registered_subject = model_cls.objects.get(
            subject_identifier=self.subject_visit.subject_identifier
        )
        return dict(
            gender=registered_subject.gender,
            dob=registered_subject.dob,
            report_datetime=self.subject_visit.report_datetime,
        )

    def get_summary(self: Any) -> list:
        opts = self.get_summary_options()
        summary = []
        for field_name in [f.name for f in self._meta.get_fields()]:
            try:
                utest_id, _ = field_name.split("_value")
            except ValueError:
                utest_id = field_name
            if reference_grp := site_reportables.get(
                get_reference_range_collection_name(self)
            ).get(utest_id):
                if value := getattr(self, field_name):
                    units = getattr(self, f"{utest_id}_units")
                    opts.update(units=units)
                    grade = reference_grp.get_grade(value, **opts)
                    if grade and grade.grade:
                        setattr(self, f"{utest_id}_grade", grade.grade)
                        setattr(self, f"{utest_id}_grade_description", grade.description)
                        summary.append(f"{field_name}: {grade.description}.")
                    elif not grade:
                        normal = reference_grp.get_normal(value, **opts)
                        if not normal:
                            summary.append(f"{field_name}: {value} {units} is abnormal")
        return summary

    def get_action_item_reason(self):
        return self.summary

    @property
    def abnormal(self: Any):
        return self.results_abnormal

    @property
    def reportable(self: Any):
        return self.results_reportable

    class Meta:
        abstract = True


class BloodResultsModelMixin(
    BloodResultsFieldsModelMixin, BloodResultsMethodsModelMixin, models.Model
):
    """For each `result` the field name or its prefix should
    match with a value in reportables.

    For example:
        field_name = creatinine_value
        reportables name: creatinine
        value_field_suffix = "_value"

        -OR-

        field_name = creatinine
        reportables name: creatinine
        value_field_suffix = None

    Requires additional attrs `subject_visit` and `requisition`
    from CrfWithRequisitionModelMixin

    """

    class Meta:
        abstract = True
